# copyright ############################### #
# This file is part of the Xsuite project.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #
import logging

import xcoll as xc
import xfields as xf

from xtrack.beam_elements import *
from xtrack.monitors import *
from xtrack.random import *
from xtrack.multisetter import MultiSetter

LOGGER = logging.getLogger(__name__)

BASE_CONFIG = {
    'XTRACK_MULTIPOLE_NO_SYNRAD': True,
    'XFIELDS_BB3D_NO_BEAMSTR': True,
    'XFIELDS_BB3D_NO_BHABHA': True,
    'XTRACK_GLOBAL_XY_LIMIT': 1.0,
}

FREEZE_ENERGY = {
    'FREEZE_VAR_delta': True,
    'FREEZE_VAR_ptau': True,
    'FREEZE_VAR_rpp': True,
    'FREEZE_VAR_rvv': True,
}

FREEZE_LONGITUDINAL = {
    **FREEZE_ENERGY,
    'FREEZE_VAR_zeta': True,
}

ONLY_XTRACK_ELEMENTS = [
    Drift,
    Multipole,
    Marker,
    ReferenceEnergyIncrease,
    Cavity,
    Elens,
    Wire,
    Solenoid,
    RFMultipole,
    DipoleEdge,
    SimpleThinBend,
    SimpleThinQuadrupole,
    LineSegmentMap,
    FirstOrderTaylorMap,
    NonLinearLens,
    # Slices
    DriftSlice,
    DriftSliceBend,
    DriftSliceOctupole,
    DriftSliceQuadrupole,
    DriftSliceSextupole,
    ThickSliceBend,
    ThickSliceOctupole,
    ThickSliceQuadrupole,
    ThickSliceSextupole,
    ThinSliceBend,
    ThinSliceBendEntry,
    ThinSliceBendExit,
    ThinSliceOctupole,
    ThinSliceQuadrupole,
    ThinSliceSextupole,
    # Transformations
    XYShift,
    ZetaShift,
    XRotation,
    SRotation,
    YRotation,
    # Apertures
    LimitEllipse,
    LimitRectEllipse,
    LimitRect,
    LimitRacetrack,
    LimitPolygon,
    LongitudinalLimitRect,
    # Monitors
    BeamPositionMonitor,
    BeamSizeMonitor,
    BeamProfileMonitor,
    LastTurnsMonitor,
    ParticlesMonitor,
]

NO_SYNRAD_ELEMENTS = [
    Bend,
    Quadrupole,
    Sextupole,
    Octupole,
    SecondOrderTaylorMap,
    Exciter,
]

NON_TRACKING_ELEMENTS = [
    RandomUniform,
    RandomExponential,
    RandomNormal,
    RandomRutherford,
    MultiSetter,
]

# These are enumerated in order specified below: the highest priority at the top
kernel_definitions = [
    ('default_only_xtrack_no_config', {
        'config': {},
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('default_only_xtrack', {
        'config': BASE_CONFIG,
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('default_only_xtrack_exact_drifts', {
        'config': {
            **BASE_CONFIG,
            'XTRACK_USE_EXACT_DRIFTS': True,
        },
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('default_only_xtrack_no_limit', {
        'config': {
            **{k: v for k, v in BASE_CONFIG.items()
                if k != 'XTRACK_GLOBAL_XY_LIMIT'}
        },
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('only_xtrack_non_tracking_kernels_no_config', {
        'config': {},
        'classes': [],
        'extra_classes': NON_TRACKING_ELEMENTS
    }),
    ('only_xtrack_non_tracking_kernels', {
        'config': BASE_CONFIG,
        'classes': [],
        'extra_classes': NON_TRACKING_ELEMENTS
    }),
    ('default_only_xtrack_backtrack', {
        'config': {**BASE_CONFIG, 'XSUITE_BACKTRACK': True},
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('default_only_xtrack_backtrack_no_limit', {
        'config': {
            **{k: v for k, v in BASE_CONFIG.items()
                if k != 'XTRACK_GLOBAL_XY_LIMIT'},
            'XSUITE_BACKTRACK': True
        },
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('only_xtrack_frozen_longitudinal', {
        'config': {**BASE_CONFIG, **FREEZE_LONGITUDINAL},
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('only_xtrack_frozen_energy', {
        'config': {**BASE_CONFIG, **FREEZE_ENERGY},
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('only_xtrack_backtrack_frozen_energy', {
        'config': {**BASE_CONFIG, **FREEZE_ENERGY, 'XSUITE_BACKTRACK': True},
        'classes': ONLY_XTRACK_ELEMENTS + NO_SYNRAD_ELEMENTS,
    }),
    ('only_xtrack_taper', {
        'config': {
            **BASE_CONFIG,
            'XTRACK_MULTIPOLE_NO_SYNRAD': False,
            'XTRACK_MULTIPOLE_TAPER': True,
            'XTRACK_DIPOLEEDGE_TAPER': True,
        },
        'classes': ONLY_XTRACK_ELEMENTS,
    }),
    ('only_xtrack_with_synrad', {
        'config': {**BASE_CONFIG, 'XTRACK_MULTIPOLE_NO_SYNRAD': False},
        'classes': ONLY_XTRACK_ELEMENTS,
    }),
    ('only_xtrack_with_synrad_kick_as_co', {
        'config': {
            **BASE_CONFIG, 'XTRACK_MULTIPOLE_NO_SYNRAD': False,
            'XTRACK_SYNRAD_KICK_SAME_AS_FIRST': True
        },
        'classes': ONLY_XTRACK_ELEMENTS,
    }),
]


# Xfields elements
DEFAULT_XF_ELEMENTS = [
    *ONLY_XTRACK_ELEMENTS,
    xf.BeamBeamBiGaussian2D,
    xf.BeamBeamBiGaussian3D,
    xf.SpaceChargeBiGaussian,
]

kernel_definitions += [
    ('default_xfields', {
        'config': BASE_CONFIG,
        'classes': [*DEFAULT_XF_ELEMENTS],
    }),
    ('default_xfields_no_config', {
        'config': {},
        'classes': [*DEFAULT_XF_ELEMENTS],
    }),
    ('default_xfields_frozen_longitudinal', {
        'config': {**BASE_CONFIG, **FREEZE_LONGITUDINAL},
        'classes': DEFAULT_XF_ELEMENTS,
    }),
    ('default_xfields_frozen_energy', {
        'config': {**BASE_CONFIG, **FREEZE_ENERGY},
        'classes': DEFAULT_XF_ELEMENTS,
    }),
]

# Xcoll elements
DEFAULT_XCOLL_ELEMENTS = [
    *ONLY_XTRACK_ELEMENTS,
    *NO_SYNRAD_ELEMENTS,
    xc.BlackAbsorber,
    xc.EverestBlock,
    xc.EverestCollimator,
    xc.EverestCrystal
]

kernel_definitions += [
    ('default_xcoll', {
        'config': BASE_CONFIG,
        'classes': DEFAULT_XCOLL_ELEMENTS,
    }),
    ('default_xcoll_no_config', {
        'config': {},
        'classes': DEFAULT_XCOLL_ELEMENTS,
    }),
    ('default_xcoll_no_limit', {
        'config': {
            **{k: v for k, v in BASE_CONFIG.items()
                if k != 'XTRACK_GLOBAL_XY_LIMIT'}
        },
        'classes': DEFAULT_XCOLL_ELEMENTS,
    }),
    ('default_xcoll_frozen_longitudinal', {
        'config': {**BASE_CONFIG, **FREEZE_LONGITUDINAL},
        'classes': DEFAULT_XCOLL_ELEMENTS,
    }),
    ('default_xcoll_frozen_energy', {
        'config': {**BASE_CONFIG, **FREEZE_ENERGY},
        'classes': DEFAULT_XCOLL_ELEMENTS,
    }),
    ('default_xcoll_backtrack', {
        'config': {**BASE_CONFIG, 'XSUITE_BACKTRACK': True},
        'classes': DEFAULT_XCOLL_ELEMENTS,
    }),
    ('default_xcoll_backtrack_no_limit', {
        'config': {
            **{k: v for k, v in BASE_CONFIG.items()
                if k != 'XTRACK_GLOBAL_XY_LIMIT'},
            'XSUITE_BACKTRACK': True
        },
        'classes': DEFAULT_XCOLL_ELEMENTS,
    }),
    ('default_xcoll_backtrack_frozen_energy', {
        'config': {**BASE_CONFIG, **FREEZE_ENERGY, 'XSUITE_BACKTRACK': True},
        'classes': DEFAULT_XCOLL_ELEMENTS,
    }),
]



# copyright ############################### #
# This file is part of the Xsuite project.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #
import json
import os
from pathlib import Path
from pprint import pformat
from typing import Iterator, Optional, Tuple

import numpy as np

import xobjects as xo
import xsuite_kernels
import xtrack as xt
from xtrack.general import _print

XSK_PREBUILT_KERNELS_LOCATION = Path(xsuite_kernels.__file__).parent / 'lib'

BEAM_ELEMENTS_INIT_DEFAULTS = {
    'Bend': {
        'length': 1.,
    },
    'Quadrupole': {
        'length': 1.,
    },
    'Solenoid': {
        'length': 1.,
    },
    'BeamBeamBiGaussian2D': {
        'other_beam_Sigma_11': 1.,
        'other_beam_Sigma_33': 1.,
        'other_beam_num_particles': 0.,
        'other_beam_q0': 1.,
        'other_beam_beta0': 1.,
    },
    'BeamBeamBiGaussian3D': {
        'slices_other_beam_zeta_center': np.array([0]),
        'slices_other_beam_num_particles': np.array([0]),
        'phi': 0.,
        'alpha': 0,
        'other_beam_q0': 1.,
        'slices_other_beam_Sigma_11': np.array([1]),
        'slices_other_beam_Sigma_12': np.array([0]),
        'slices_other_beam_Sigma_22': np.array([0]),
        'slices_other_beam_Sigma_33': np.array([1]),
        'slices_other_beam_Sigma_34': np.array([0]),
        'slices_other_beam_Sigma_44': np.array([0]),
    },
    'LimitPolygon': {
        'x_vertices': np.array([0, 1, 1, 0]),
        'y_vertices': np.array([0, 0, 1, 1]),
    },
    'BeamProfileMonitor': {
        'range': 1,
    },
    'LastTurnsMonitor': {
        'n_last_turns': 1,
        'num_particles': 1,
    },
    'ParticlesMonitor': {
        'num_particles': 1,
        'start_at_turn': 0,
        'stop_at_turn': 1,
    },
    'Exciter': {
        'samples': [0],
    },
    'SpaceChargeBiGaussian': {
        'longitudinal_profile': {
            '__class__': 'LongitudinalProfileQGaussian',
            'number_of_particles': 1,
            'sigma_z': 0,
        }
    },
}

# SpaceChargeBiGaussian is not included for now (different issues -
# circular import, incompatible compilation flags)


def get_element_class_by_name(name: str) -> type:
    try:
        from xfields import element_classes as xf_element_classes
    except ModuleNotFoundError:
        xf_element_classes = ()

    try:
        from xcoll import element_classes as xc_element_classes
    except ModuleNotFoundError:
        xc_element_classes = ()

    xt_multisetter = (xt.MultiSetter, )

    element_classes = (xt.element_classes + xt.rng_classes
                       + xt.monitor_classes + xt_multisetter
                       + xf_element_classes + xc_element_classes)

    for cls in element_classes:
        if cls.__name__ == name:
            return cls

    raise ValueError(f'No element class with name {name} available.')


def save_kernel_metadata(
        module_name: str,
        config: dict,
        kernel_element_classes,
        location,
):
    location = Path(location)
    out_file = location / f'{module_name}.json'

    try:
        import xfields
        xf_version = xfields.__version__
    except ModuleNotFoundError:
        xf_version = None

    try:
        import xcoll
        xc_version = xcoll.__version__
    except ModuleNotFoundError:
        xc_version = None

    kernel_metadata = {
        'config': config.data,
        'classes': [cls._DressingClass.__name__ for cls in kernel_element_classes],
        'versions': {
            'xtrack': xt.__version__,
            'xfields': xf_version,
            'xcoll': xc_version,
            'xobjects': xo.__version__,
        }
    }

    with out_file.open('w') as fd:
        json.dump(kernel_metadata, fd, indent=4)


def enumerate_kernels() -> Iterator[Tuple[str, dict]]:
    """
    Iterate over the prebuilt kernels compatible with the current version of
    xsuite. The first element of the tuple is the name of the kernel module
    and the second is a dictionary with the kernel metadata.
    """
    from xsuite_kernels.kernel_definitions import kernel_definitions
    for kernel_name, _ in kernel_definitions:
        metadata_file = XSK_PREBUILT_KERNELS_LOCATION / f'{kernel_name}.json'

        if not metadata_file.exists():
            continue

        with metadata_file.open('r') as fd:
            kernel_metadata = json.load(fd)

        try:
            import xfields
            xf_version = xfields.__version__
        except ModuleNotFoundError:
            xf_version = None

        try:
            import xcoll
            xc_version = xcoll.__version__
        except ModuleNotFoundError:
            xc_version = None

        if kernel_metadata['versions']['xtrack'] != xt.__version__:
            continue

        if kernel_metadata['versions']['xobjects'] != xo.__version__:
            continue

        if (kernel_metadata['versions']['xfields'] != xf_version
                and xf_version is not None):
            continue

        if (kernel_metadata['versions']['xcoll'] != xc_version
                and xc_version is not None):
            continue

        yield metadata_file.stem, kernel_metadata


def get_suitable_kernel(
        config: dict,
        line_element_classes,
        verbose=False,
) -> Optional[Tuple[str, list]]:
    """
    Given a configuration and a list of element classes, return a tuple with
    the name of a suitable prebuilt kernel module together with the list of
    element classes that were used to build it. Set `verbose` to True, to
    obtain a justification of the choice (or lack thereof) on standard output.
    """
    _print.suppress = False
    env_var = os.environ.get("XSUITE_PREBUILT_KERNELS")
    if env_var and env_var == '0':
        if verbose:
            _print('Skipping the search for a suitable kernel, as the '
                   'environment variable XSUITE_PREBUILT_KERNELS == "0".')
        return

    requested_class_names = [
        cls._DressingClass.__name__ for cls in line_element_classes
    ]

    for module_name, kernel_metadata in enumerate_kernels():
        if verbose:
            _print(f"==> Considering the precompiled kernel `{module_name}`...")

        available_classes_names = kernel_metadata['classes']
        if kernel_metadata['config'] != config:
            if verbose:
                lhs = kernel_metadata['config']
                rhs = config
                config_diff = {kk: (lhs.get(kk), rhs.get(kk))
                               for kk in set(lhs.keys()) | set(rhs.keys())
                               if lhs.get(kk) != rhs.get(kk)}
                _print(f'The kernel `{module_name}` is unsuitable. Its config '
                      f'(left) and the requested one (right) differ at the '
                      f'following keys:\n'
                      f'{pformat(config_diff)}')
                _print(f'Skipping class compatibility check for `{module_name}`.')

            continue

        if verbose:
            _print(f'The kernel `{module_name}` has the right config.')

        if set(requested_class_names) <= set(available_classes_names):
            available_classes = [
                get_element_class_by_name(class_name)
                for class_name in available_classes_names
            ]
            _print(f'Found suitable prebuilt kernel `{module_name}`.')
            return module_name, available_classes
        elif verbose:
            class_diff = set(requested_class_names) - set(available_classes_names)
            _print(f'The kernel `{module_name}` is unsuitable. It does not '
                  f'provide the following requested classes: '
                  f'{", ".join(class_diff)}.')

    if verbose:
        _print('==> No suitable precompiled kernel found.')


def regenerate_kernels(kernels=None, location=XSK_PREBUILT_KERNELS_LOCATION):
    """
    Use the kernel definitions in the `kernel_definitions.py` file to
    regenerate kernel shared objects using the current version of xsuite.
    """
    if kernels is not None and (
    isinstance(kernels, str) or not hasattr(kernels, '__iter__')):
        kernels = [kernels]

    location = Path(location)

    # Delete existing kernels to avoid accidentally loading in existing C code
    clear_kernels(kernels=kernels, location=location)

    import xpart as xp
    from xsuite_kernels.kernel_definitions import kernel_definitions
    try:
        import xcoll as xc
        BEAM_ELEMENTS_INIT_DEFAULTS['EverestBlock'] = {
                'material': xc.materials.Silicon,
                'use_prebuilt_kernels': False
            }
        BEAM_ELEMENTS_INIT_DEFAULTS['EverestCollimator'] = {
                'material': xc.materials.Silicon,
                'use_prebuilt_kernels': False
            }
        BEAM_ELEMENTS_INIT_DEFAULTS['EverestCrystal'] = {
                'material': xc.materials.SiliconCrystal,
                'use_prebuilt_kernels': False
            }
    except ImportError:
        pass

    kernels_to_build = []
    for module_name, metadata in kernel_definitions:
        if kernels is not None and module_name not in kernels:
            continue
        kernels_to_build.append((module_name, metadata))

    for idx, (module_name, metadata) in enumerate(kernels_to_build):
        _print(f'[{idx + 1}/{len(kernels_to_build)}] Building `{module_name}`...')

        config = metadata['config']
        element_classes = metadata['classes']
        extra_classes = metadata.get('extra_classes', [])

        elements = []
        for cls in element_classes:
            if cls.__name__ in BEAM_ELEMENTS_INIT_DEFAULTS:
                element = cls(**BEAM_ELEMENTS_INIT_DEFAULTS[cls.__name__])
            else:
                element = cls()
            elements.append(element)

        line = xt.Line(elements=elements)
        tracker = xt.Tracker(line=line, compile=False, _prebuilding_kernels=True)
        tracker.config.clear()
        tracker.config.update(config)

        # Get all kernels in the elements
        extra_kernels = {}
        extra_classes.append(xp.Particles)
        extra_classes = [getattr(el, '_XoStruct', el) for el in extra_classes]
        all_classes = tracker._tracker_data_base.kernel_element_classes + extra_classes
        for el in all_classes:
            extra_kernels.update(el._kernels)

        # TODO: Add any other kernels that are defined in the context
        #       Need to add the source etc
        # kernel_descriptions.update(tracker._context.kernels)

        tracker._build_kernel(
            module_name=module_name,
            containing_dir=location,
            compile='force',
            extra_classes=extra_classes,
            extra_kernels=extra_kernels,
        )

        all_classes = [cls for cls in all_classes
                       if cls.__name__ != 'ParticlesData']
        save_kernel_metadata(
            module_name=module_name,
            config=tracker.config,
            kernel_element_classes=all_classes,
            location=location,
        )


def clear_kernels(kernels=None, verbose=False, location=XSK_PREBUILT_KERNELS_LOCATION):
    if kernels is not None and (
            isinstance(kernels, str) or not hasattr(kernels, '__iter__')):
        kernels = [kernels]

    location = Path(location)

    for file in location.iterdir():
        if file.name.startswith('_'):
            continue
        if file.suffix not in ('.c', '.so', '.json'):
            continue
        if kernels is not None and file.stem.split('.')[0] not in kernels:
            continue
        file.unlink()

        if verbose:
            print(f'Removed `{file}`.')


if __name__ == '__main__':
    regenerate_kernels()




# copyright ############################### #
# This file is part of the Xsuite project.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #
import argparse
from importlib.metadata import version

from xsuite_kernels.prebuild_kernels import (
    clear_kernels, regenerate_kernels, XSK_PREBUILT_KERNELS_LOCATION,
)


def regenerate_command(args):
    regenerate_kernels()
    print('Kernels regenerated.')


def clean_command(args):
    clear_kernels(verbose=args.verbose)
    print('Cleaned kernels.')


def info_command(args):
    version_str = version("xsuite_kernels")
    print(f'Xsuite kernels version {version_str}')
    print(f'Kernels location: {XSK_PREBUILT_KERNELS_LOCATION}')


def main():
    parser = argparse.ArgumentParser(
        prog='xsuite-kernels',
        description='Regenerate and clean precompiled kernels for Xsuite.',
    )
    subparsers = parser.add_subparsers(
        title='commands',
        required=True,
    )

    # `regenerate` commend
    regenerate_parser = subparsers.add_parser(
        'regenerate',
        description='Regenerate the kernels.'
    )
    regenerate_parser.set_defaults(func=regenerate_command)

    # `clean` command
    clean_parser = subparsers.add_parser(
        'clean',
        description='Clean the kernel directory.'
    )
    clean_parser.add_argument(
        '-v', '--verbose',
        help='list the files being deleted',
        action='store_true',
    )
    clean_parser.set_defaults(func=clean_command)

    # `info` command
    info_parser = subparsers.add_parser(
        'info',
        description='Print additional info.',
    )
    info_parser.set_defaults(func=info_command)

    # handle the right action
    args = parser.parse_args()
    args.func(args)



# copyright ############################### #
# This file is part of the Xsuite project.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #
from .prebuild_kernels import get_suitable_kernel, XSK_PREBUILT_KERNELS_LOCATION
